#!/usr/bin/env python

"""
RSVP instruction classes.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import Label
from micropython.data import Attr, Const, Instance

def raw(code):
    new_code = []
    for item in code:
        if isinstance(item, Attr):
            new_code.append((item.context and item.context.location, item.value and item.value.location))
        else:
            new_code.append(item)
    return new_code

def name(attr):
    if isinstance(attr, Attr):
        return attr.name
    else:
        return attr

class Instruction:

    "A generic instruction."

    stack_usage = 0
    stack_access = 0

    def __init__(self, attr=None):
        self.attr = attr
        self.operands = [StackLoad(-1 - n) for n in range(0, self.stack_access)]

    def fix_stack(self, level):
        for operand in self.operands:
            operand.fix_stack(level)

    def show_operands(self):
        return self.operands and (" %s" % self.operands) or ""

    def __repr__(self):
        if self.attr is not None:
            return "%s(%r)%s" % (self.__class__.__name__, self.attr, self.show_operands())
        else:
            return "%s()%s" % (self.__class__.__name__, self.show_operands())

class StackRelativeInstruction(Instruction):

    "An instruction operating on the local value stack."

    def __repr__(self):
        return "%s(%r)%s" % (self.__class__.__name__, self.get_operand(), self.show_operands())

    def get_operand(self):
        return self.attr.position

SR = StackRelativeInstruction

class AddressRelativeInstruction(Instruction):

    "An instruction accessing an object's attribute."

    def __repr__(self):
        position = self.get_operand()
        if position is not None:
            return "%s(%r)%s # %s" % (self.__class__.__name__, position, self.show_operands(), name(self.attr))
        else:
            return "%s(%r)%s" % (self.__class__.__name__, self.show_operands(), name(self.attr))

    def get_operand(self):
        return self.attr.position

AR = AddressRelativeInstruction

class AddressInstruction(Instruction):

    "An instruction loading an address directly."

    def __repr__(self):
        location, position, result = self.get_operands()
        if location is not None:
            return "%s(%r)%s # %r, %r (%s)" % (self.__class__.__name__, result, self.show_operands(), location, position, name(self.attr))
        elif result is not None:
            return "%s(%r)%s # %s" % (self.__class__.__name__, result, self.show_operands(), name(self.attr))
        else:
            return "%s(...)%s # %s" % (self.__class__.__name__, self.show_operands(), name(self.attr))

    def get_operands(self):
        if isinstance(self.attr, Attr):
            position = self.attr.position
            location = self.attr.parent.location

            # NOTE: Unpositioned attributes are handled here.

            if location is not None and position is not None:
                result = location + position + 1
            else:
                location = self.attr.parent.name
                position = self.attr.name
                result = None
            return location, position, result
        elif isinstance(self.attr, Label):
            return None, None, self.attr.location
        else:
            return None, None, self.attr.location

    def get_operand(self):
        return self.get_operands()[-1]

Address = AddressInstruction

class ImmediateInstruction(Instruction):

    "An instruction employing a constant."

    def __repr__(self):
        return "%s(%r)%s" % (self.__class__.__name__, self.attr, self.show_operands())

    def get_operand(self):
        return self.attr

Immediate = ImmediateInstruction

# Mix-in classes for stack effects.

class StackAdd:

    """
    Indicate that the stack must grow to accommodate the result of this
    instruction.
    """

    stack_usage = 1

class StackRemove:

    "Indicate that the stack must shrink as an effect of this instruction."

    stack_usage = -1
    stack_access = 1

class StackRemove2:

    "Indicate that the stack must shrink as an effect of this instruction."

    stack_usage = -2
    stack_access = 2

# Instructions operating on the value stack.

class Duplicate(StackAdd, Instruction):         "Duplicate the top of the stack."
class ResetStack(Instruction):                  "Pop entries from the stack."

# Access to stored constant data.

class LoadConst(StackAdd, Address):             "Load the constant, class, function, module from the specified location."

# Access within an invocation frame.

class LoadName(StackAdd, SR):                   "Load the object from the given local attribute/variable."
class StoreName(StackRemove, SR):               "Store the object in the given local attribute/variable."
class LoadTemp(Immediate):                      "Load the object from the given temporary location."
class StoreTemp(StackRemove, Immediate):        "Store the object in the given temporary location."

# Access to address-relative data.

class MakeObject(StackAdd, Instruction):        "Make a new object. There isn't a complementary DropObject."
class LoadAttr(AR):                             "Load the object from the given attribute."
class StoreAttr(StackRemove2, AR):              "Store an object in the given attribute."
class LoadAttrIndex(Immediate):                 "Load the object for the attribute with the given index."
class StoreAttrIndex(StackRemove2, Immediate):  "Store an object in the attribute with the given index."
class LoadAddress(StackAdd, Address):           "Load the object from the given fixed attribute address."
class StoreAddress(StackRemove, Address):       "Store an object in the given fixed attribute address."

# Access to invocation frames in preparation.

class MakeFrame(Instruction):                   "Make a new invocation frame."
class ReserveFrame(Immediate):                  "Reserve the given number of entries for the invocation frame."
class DropFrame(Instruction):                   "Drop an invocation frame."
class StoreFrame(StackRemove, Immediate):       "Store an argument at the given frame location."
class StoreFrameIndex(StackRemove, Immediate):  "Store an argument for the parameter with the given index."
class CheckFrame(Instruction):                  "Check the invocation frame for the target."
class JumpWithFrame(Instruction):               "Jump, adopting the invocation frame, to the callable found on the stack."

# Invocation-related instructions.

class Jump(Address):                            "Jump unconditionally."
class JumpIfFalse(Address):                     "Jump if the last evaluation gave a false result."
class JumpIfTrue(Address):                      "Jump if the last evaluation gave a true result."
class LoadCallable(Instruction):                "Load the target of an invocation."
class LoadContext(Instruction):                 "Load the context of an invocation."
class CheckContext(Instruction):                """Check the context of an invocation against the target,
                                                potentially discarding the context."""
class CheckSelf(Instruction):                   "Check the first argument of an invocation against the target."
class RaiseException(Instruction):              "Raise an exception."
class Return(Instruction):                      "Return a value from a subprogram."
class CheckException(Instruction):              "Check the raised exception against another."

# General instructions.

class TestIdentity(Instruction):                "Test whether the two topmost stack values are identical."
class TestIdentityAddress(Address):             "Test whether the topmost stack value is identical to the given address."

# Internal stack operations for instructions.

class StackLoad:

    "Load a value from the stack."

    def __init__(self, n):
        self.n = n
        self.level = None

    def fix_stack(self, level):
        self.level = self.n + level

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.n)

# vim: tabstop=4 expandtab shiftwidth=4
